use super::{
    compiler::Compiler,
    expr::{Expr, ExprOp, Variable},
    opcode::OpCode,
    parser::parse,
    runtime::execute,
    value::{Value, ValueOrRef},
    Addr, Error, Register,
};
use crate::Transaction;
use std::{collections::HashMap, rc::Rc};

#[derive(Debug, Clone)]
pub struct Filter {
    /// Query
    expr: Option<Expr>,
    /// List of external variables
    variables: Vec<Variable>,
    /// Unoptimized opcodes
    dbg_opcodes: Vec<OpCode>,
    /// Opcodes with human readable labels
    opcodes: Vec<OpCode>,
    /// The list of opcodes that make up the program, the labels has been converted into addresses
    opcodes_to_execute: Vec<OpCode>,
    /// The address of the initial opcode
    initial_opcode_addr: Addr,
    /// The state of the register
    pub(crate) initial_register: HashMap<Register, ValueOrRef>,
}

impl Filter {
    pub fn new(code: &str) -> Result<Self, Error> {
        let expr = parse(code)?;
        let opcodes = expr.as_ref().map_or_else(
            || {
                Ok(vec![
                    OpCode::LOAD(0.into(), Rc::new(true.into())),
                    OpCode::HLT(0.into()),
                ])
            },
            |expr| Compiler::new(expr).compile(),
        )?;

        let instance = Self {
            expr,
            dbg_opcodes: opcodes.clone(),
            opcodes: opcodes.clone(),
            initial_opcode_addr: 0.into(),
            variables: opcodes
                .iter()
                .filter_map(|x| match x {
                    OpCode::LOAD_EXTERNAL(_, name) => Some(name.clone()),
                    _ => None,
                })
                .collect(),
            opcodes_to_execute: Compiler::resolve_label_to_addr(opcodes)?,
            initial_register: HashMap::new(),
        };
        Ok(instance.optimize())
    }

    /// Optimizes the program
    pub fn optimize(mut self) -> Self {
        let mut new_opcodes = self.opcodes;
        loop {
            let (new_opcodes_, has_changed) = Compiler::optimize(new_opcodes);
            new_opcodes = new_opcodes_;
            if !has_changed {
                break;
            }
        }

        self.opcodes = new_opcodes;
        self.initial_register.clear();
        self.opcodes_to_execute = Compiler::resolve_label_to_addr(self.opcodes.clone()).unwrap();

        for (pos, opcode) in self.opcodes.iter().enumerate() {
            // Execute all the LOAD opcodes until we find a non-LOAD opcode
            // and store the initial state of the registers
            match opcode {
                OpCode::LOAD(r, v) => {
                    self.initial_register
                        .insert(*r, ValueOrRef::WeakRef(Rc::downgrade(v)));
                }
                _ => {
                    self.initial_opcode_addr = pos.into();
                    break;
                }
            }
        }

        self
    }

    /// Returns a human readable version of the compiled program (generated op-codes)
    pub fn debug(&self) -> String {
        self.dbg_opcodes
            .iter()
            .map(|x| match x {
                OpCode::HLT(_) | OpCode::LABEL(_) => x.to_string(),
                x => format!("\t{}", x.to_string()),
            })
            .collect::<Vec<_>>()
            .join("\n")
    }

    pub fn dump(&self) -> String {
        self.opcodes_to_execute
            .iter()
            .enumerate()
            .map(|(pos, opcode)| format!("{}: {}", pos, opcode.to_string()))
            .collect::<Vec<_>>()
            .join("\n")
    }

    pub fn execute(
        &self,
        external_variables: &HashMap<Variable, ValueOrRef>,
    ) -> Result<Value, Error> {
        execute(
            external_variables,
            &self.opcodes_to_execute,
            self.initial_opcode_addr,
            self.initial_register.clone(),
        )
    }
}
