use super::{
    expr::Variable,
    opcode::OpCode,
    value::{Value, ValueOrRef},
    Addr, Error, Register,
};
use crate::Transaction;
use num::CheckedAdd;
use std::{collections::HashMap, ops::Deref};

macro_rules! get {
    ($r:expr,$pos:expr) => {
        ($r.get($pos)
            .ok_or_else(|| Error::RegisterOutOfBoundaries($pos.to_owned()))?)
    };
}

macro_rules! remove {
    ($r:expr,$pos:expr) => {
        ($r.remove($pos)
            .ok_or_else(|| Error::RegisterOutOfBoundaries($pos.to_owned()))?)
    };
}

macro_rules! set {
    ($r:expr, $pos:expr, $new_value:expr) => {
        (if let Some(previous_value) = $r.get_mut($pos) {
            *previous_value = $new_value;
        } else {
            $r.insert(($pos).to_owned(), $new_value);
        })
    };
}

#[inline]
/// Execute the program
///
/// Executes the program passing the external variables and the initial registers.
///
/// The external variables is a reference to a hashmap that contains the
/// external variables, if the program attempts to load a variable that does not
/// exists will return an error immediately.
///
/// The initial registers is a hashmap that contains the initial state of the
/// registers, this is a tool available for the compiler to execute the initial
/// LOAD, if it makes sense, and pass over the initial state of the registers to
/// the program, avoiding executing the same opcodes over and over
pub fn execute(
    external_variables: &HashMap<Variable, ValueOrRef>,
    code: &[OpCode],
    start_at: Addr,
    initial_registers: HashMap<Register, ValueOrRef>,
) -> Result<Value, Error> {
    let mut execution: Addr = start_at;
    let mut registers = initial_registers;

    loop {
        match code.get(*execution).ok_or(Error::OutOfBoundaries)? {
            OpCode::LABEL(_addr) => {}
            OpCode::LOAD_EXTERNAL(dst, name) => {
                let value = external_variables
                    .get(name)
                    .ok_or(Error::VariableNotFound((*name).join(".")))?;

                set!(registers, dst, value.clone());
            }
            OpCode::LOAD(dst, ref val) => set!(registers, dst, val.into()),
            OpCode::CPY(dst, reg2) => {
                let value = get!(registers, reg2).clone();
                set!(registers, dst, value);
            }
            OpCode::MOV(dst, reg2) => {
                let previous_value = remove!(registers, reg2);
                set!(registers, dst, previous_value);
            }
            OpCode::ADD(dst, a, b) => {
                let new_value = get!(registers, a)
                    .checked_add(get!(registers, b))
                    .ok_or(Error::Overflow)?
                    .into();
                set!(registers, dst, new_value);
            }
            OpCode::SUB(dst, reg1, reg2) => {
                let num1 = get!(registers, reg1).as_number()?;
                let num2 = get!(registers, reg2).as_number()?;
                set!(
                    registers,
                    dst,
                    num1.checked_sub(num2)
                        .map(|number| Value::Number(number).into())
                        .ok_or(Error::Overflow)?
                );
            }
            OpCode::AND(dst, reg2, reg3) => {
                let new_value = Value::Bool(
                    get!(registers, reg2).as_boolean()? && get!(registers, reg3).as_boolean()?,
                )
                .into();
                set!(registers, dst, new_value);
            }
            OpCode::MOD(dst, reg1, reg2) => {
                let num1 = get!(registers, reg1).as_number()?;
                let num2 = get!(registers, reg2).as_number()?;
                set!(
                    registers,
                    dst,
                    num1.checked_rem(num2)
                        .map(|number| Value::Number(number).into())
                        .ok_or(Error::Overflow)?
                );
            }
            OpCode::NEG(dst, reg1) => {
                let num1 = get!(registers, reg1).as_number()?;
                set!(
                    registers,
                    dst,
                    num1.checked_neg()
                        .map(|number| Value::Number(number).into())
                        .ok_or(Error::Overflow)?
                );
            }
            OpCode::GE(dst, reg1, reg2) => {
                let new_value = Value::Bool(get!(registers, reg1) >= get!(registers, reg2)).into();
                set!(registers, dst, new_value);
            }
            OpCode::GT(dst, reg1, reg2) => {
                let new_value = Value::Bool(get!(registers, reg1) > get!(registers, reg2)).into();
                set!(registers, dst, new_value);
            }
            OpCode::LE(dst, reg1, reg2) => {
                let new_value = Value::Bool(get!(registers, reg1) <= get!(registers, reg2)).into();
                set!(registers, dst, new_value);
            }
            OpCode::LT(dst, reg1, reg2) => {
                let new_value = Value::Bool(get!(registers, reg1) < get!(registers, reg2)).into();
                set!(registers, dst, new_value);
            }
            OpCode::EQ(dst, reg1, reg2) => {
                let new_value = Value::Bool(get!(registers, reg1) == get!(registers, reg2)).into();
                set!(registers, dst, new_value);
            }
            OpCode::NE(dst, reg1, reg2) => {
                let new_value = Value::Bool(get!(registers, reg1) != get!(registers, reg2)).into();
                set!(registers, dst, new_value);
            }
            OpCode::DIV(dst, reg1, reg2) => {
                let num1 = get!(registers, reg1).as_number()?;
                let num2 = get!(registers, reg2).as_number()?;
                set!(
                    registers,
                    dst,
                    num1.checked_div(num2)
                        .map(|number| Value::Number(number).into())
                        .ok_or(Error::Overflow)?
                );
            }
            OpCode::MUL(dst, reg1, reg2) => {
                let num1 = get!(registers, reg1).as_number()?;
                let num2 = get!(registers, reg2).as_number()?;
                set!(
                    registers,
                    dst,
                    num1.checked_mul(num2)
                        .map(|number| Value::Number(number).into())
                        .ok_or(Error::Overflow)?
                );
            }
            OpCode::OR(dst1, reg2, reg3) => {
                todo!()
            }
            OpCode::XOR(dst, reg2, reg3) => {
                todo!()
            }
            OpCode::NOT(dst, reg2) => {
                let new_value = Value::Bool(!get!(registers, reg2).as_boolean()?).into();
                set!(registers, dst, new_value);
            }
            OpCode::JMP(addr) => {
                execution.jump_to(addr);
                continue;
            }
            OpCode::JEQ(reg, addr) => {
                if get!(registers, reg).as_boolean()? {
                    execution.jump_to(addr);
                    continue;
                }
            }
            OpCode::JNE(reg, addr) => {
                if !get!(registers, reg).as_boolean()? {
                    execution.jump_to(addr);
                    continue;
                }
            }
            OpCode::HLT(return_register) => {
                return registers
                    .remove(return_register)
                    .map(|x| (*x).clone())
                    .ok_or_else(|| Error::EmptyRegister(return_register.to_owned()));
            }
            x => panic!("Unimplemented opcode: {:?}", x),
        }

        execution.next();
    }
}
