use crate::{
    filter_expr::{
        expr::{Expr, Variable},
        Error,
    },
    payment::PaymentTo,
    AccountId, Amount, Asset, PaymentFrom, RevId, Status, Tag, TxId, Type,
};
use chrono::{DateTime, Utc};
use num::CheckedAdd;
use std::{
    ops::{Add, Deref},
    rc::{Rc, Weak},
};

#[derive(Clone, Debug, PartialEq, PartialOrd)]
pub enum Value {
    /// Not a value
    Nil,
    /// Amount
    Amount(Amount),
    /// Asset
    Asset(Asset),
    /// Account Type
    AccountId(AccountId),
    /// Transaction ID
    TxId(TxId),
    /// Type
    Type(Type),
    /// Revision Id
    RevId(RevId),
    /// Status
    Status(Status),
    /// Tag
    Tag(Tag),
    /// Datetime
    Date(DateTime<Utc>),
    /// String
    String(String),
    /// Integer number
    Number(i128),
    /// PaymentFrom
    From(Vec<PaymentFrom>),
    /// Payment Recipient
    To(Vec<PaymentTo>),
    /// Tags
    Tags(Vec<Tag>),
    /// Bool values
    Bool(bool),
    /// A dynamic variable name
    Variable(Variable),
}

impl From<&Variable> for Value {
    fn from(value: &Variable) -> Self {
        Self::Variable(value.clone())
    }
}

impl From<bool> for Value {
    fn from(x: bool) -> Self {
        Value::Bool(x)
    }
}

impl From<i128> for Value {
    fn from(x: i128) -> Self {
        Value::Number(x)
    }
}

impl Value {
    /// Attempts to convert the value as a boolean if possible or returns an error.
    pub fn as_boolean(&self) -> Result<bool, Error> {
        match self {
            Value::Bool(b) => Ok(*b),
            Value::Nil => Ok(false),
            Value::Number(x) => Ok(*x > 0),
            /// TODO: Throw an error . Add types for convertion errors
            _ => panic!("Cannot convert to bool"),
        }
    }

    /// Checks if the value is TRUE, any other value will return FALSE.
    pub fn is_true(&self) -> bool {
        match self {
            Value::Bool(b) => *b,
            _ => false,
        }
    }

    /// Checks if the value is FALSE, any other value will return FALSE.
    pub fn is_false(&self) -> bool {
        match self {
            Value::Bool(b) => *b == false,
            _ => false,
        }
    }

    pub fn as_number(&self) -> Result<i128, Error> {
        match self {
            Value::Number(x) => Ok(*x),
            Value::String(x) => x.parse().map_err(|_| Error::NotANumber),
            _ => Err(Error::NotANumber),
        }
    }
}

impl Add for Value {
    type Output = Self;

    fn add(self, other: Self) -> Self {
        Value::Number(
            self.as_number()
                .expect("Not a number")
                .checked_add(other.as_number().expect("Not a number"))
                .expect("Overflow"),
        )
    }
}

impl CheckedAdd for Value {
    fn checked_add(&self, other: &Self) -> Option<Self> {
        self.as_number()
            .ok()?
            .checked_add(other.as_number().ok()?)
            .map(Value::Number)
    }
}

#[derive(Debug, Clone)]
/// Value or reference to a value.
///
/// This is a sorf of attempt to avoid cloning from the source code to the
/// registers, specially when the value is being defined in the source code,
/// instead
/// a Weak reference is being used to avoid cloning the value.
pub enum ValueOrRef {
    WeakRef(Weak<Value>),
    Value(Value),
}

impl Deref for ValueOrRef {
    type Target = Value;

    fn deref(&self) -> &Self::Target {
        match self {
            ValueOrRef::WeakRef(value) => {
                if let Some(rc_value) = value.upgrade() {
                    unsafe { &*(Rc::into_raw(rc_value) as *const Value) }
                } else {
                    &Value::Nil
                }
            }
            ValueOrRef::Value(v) => v,
        }
    }
}

impl PartialEq for ValueOrRef {
    fn eq(&self, other: &Self) -> bool {
        self.deref() == other.deref()
    }
}

impl PartialOrd for ValueOrRef {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        self.partial_cmp(other)
    }
}

impl From<Rc<Value>> for ValueOrRef {
    fn from(value: Rc<Value>) -> Self {
        ValueOrRef::WeakRef(Rc::downgrade(&value))
    }
}

impl From<&Rc<Value>> for ValueOrRef {
    fn from(value: &Rc<Value>) -> Self {
        ValueOrRef::WeakRef(Rc::downgrade(&value))
    }
}

impl From<Value> for ValueOrRef {
    fn from(value: Value) -> Self {
        ValueOrRef::Value(value)
    }
}
